@@ -13,7 +13,7 @@
  * This Copyright notice must be kept intact and distributed with all sources.
  */
 /*
- *  $Id: unionfs.h,v 1.104 2005/03/24 03:26:46 dquigley Exp $
+ *  $Id: unionfs.h,v 1.119 2005/05/12 20:57:52 dquigley Exp $
  */
 
 #ifndef __UNIONFS_H_
@@ -21,7 +21,7 @@
 
 #ifdef __KERNEL__
 
-#if (!defined(UNIONFS_UNSUPPORTED)) &&(LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)) || (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0) && LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9))
+#if (!defined(UNIONFS_UNSUPPORTED)) &&(LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) && LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9))
 #warning You are compiling Unionfs on an unsupported kernel version.
 #warning To compile Unionfs, you will need to define UNIONFS_UNSUPPORTED.
 #warning Try adding: EXTRACFLAGS=-DUNIONFS_UNSUPPORTED to fistdev.mk
@@ -36,7 +36,7 @@
 #endif
 
 /* fist file systems superblock magic */
-#define UNIONFS_SUPER_MAGIC 0xf15f
+#define UNIONFS_SUPER_MAGIC 0xf15f083d
 
 /* unionfs root inode number */
 #define UNIONFS_ROOT_INO     1
@@ -44,10 +44,6 @@
 /* Mount time flags */
 #define MOUNT_FLAG(sb)     (stopd(sb)->usi_mount_flag)
 
-/* maximum number of parent directories to create. */
-/* XXX: This shouldn't be so small, and it should be kmalloced. */
-#define MAX_DIR_CREATE  10
-
 /* What type of flags to use for kmalloc, etc. */
 #define GFP_UNIONFS GFP_KERNEL
 
@@ -149,51 +145,61 @@ extern struct export_operations unionfs_export_ops;
 #define RDCACHE_JIFFIES 5*HZ
 /* unionfs inode data in memory */
 struct unionfs_inode_info {
-    inode_t ** uii_inode;
-    int b_start;
-    int b_end;
-    atomic_t uii_generation;
-    int uii_stale;
-    /* Stuff for readdir over NFS. */
-    spinlock_t uii_rdlock;
-    struct list_head uii_readdircache;
-    int uii_rdversion;
-    int uii_rdcount;
-    int uii_hashsize;
+	struct inode ** uii_inode;
+	int b_start;
+	int b_end;
+	atomic_t uii_generation;
+	int uii_stale;
+	/* Stuff for readdir over NFS. */
+	spinlock_t uii_rdlock;
+	struct list_head uii_readdircache;
+	int uii_rdcount;
+	int uii_hashsize;
 	int uii_cookie;
 };
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
+struct unionfs_inode_container {
+	struct unionfs_inode_info info;
+	struct inode vfs_inode;
+};
+#endif
 
 /* unionfs dentry data in memory */
 struct unionfs_dentry_info {
-    struct multilock udi_lock;
-    dentry_t ** udi_dentry;
-    int udi_bstart;
-    int udi_bend;
-    int udi_bcount;
-    atomic_t udi_generation;
+	struct multilock udi_lock;
+	struct dentry ** udi_dentry;
+	int udi_bstart;
+	int udi_bend;
+	int udi_bopaque;
+	int udi_bcount;
+	atomic_t udi_generation;
 };
 
 /* unionfs super-block data in memory */
 struct unionfs_sb_info {
-    struct super_block *usi_thissb;
-    super_block_t **usi_sb;
-    atomic_t *usi_sbcount;
-    vfs_mount_t **usi_hidden_mnt;
-    int *usi_branchperms;
+	struct super_block *usi_thissb;
+	struct super_block **usi_sb;
+	atomic_t *usi_sbcount;
+	struct vfsmount **usi_hidden_mnt;
+	int *usi_branchperms;
+	int b_start;
+	int b_end;
 
-    int b_start;
-    int b_end;
+	uid_t copyupuid;
+	gid_t copyupgid;
+	umode_t copyupmode;
 
-    uid_t copyupuid;
-    gid_t copyupgid;
-    umode_t copyupmode;
-
-    atomic_t usi_generation;
-    unsigned long usi_mount_flag;
+	atomic_t usi_generation;
+	unsigned long usi_mount_flag;
 
+	int usi_diropaque;	
+	
+	struct file* usi_forwardmap;
+	struct file** usi_reversemaps;
+	
 #ifdef SPLIT_VIEW_CACHES
-    struct super_block *usi_primary;
-    struct list_head usi_altsupers;
+	struct super_block *usi_primary;
+	struct list_head usi_altsupers;
 #endif
 };
 
@@ -202,21 +208,25 @@ struct unionfs_sb_info {
  * to compare with entries on right branch
  */
 struct filldir_node {
-    struct list_head file_list;   // list for directory entries
-    char *name;                   // name entry
-    int hash;                     // name hash
-    int namelen;                  // name len since name is not 0 terminated
-    int bindex;                   // we can check for duplicate whiteouts and files in the same branch in order to return -EIO.
-    int whiteout;		  // is this a whiteout entry?
+	struct list_head file_list;   // list for directory entries
+	char *name;                   // name entry
+	int hash;                     // name hash
+	int namelen;                  // name len since name is not 0 terminated
+	int bindex;                   // we can check for duplicate whiteouts and files in the same branch in order to return -EIO.
+	int whiteout;		  // is this a whiteout entry?
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
 #define DNAME_INLINE_LEN_MIN DNAME_INLINE_LEN
 #endif
-    char iname[DNAME_INLINE_LEN_MIN]; // Inline name, so we don't need to separately kmalloc small ones
+	char iname[DNAME_INLINE_LEN_MIN]; // Inline name, so we don't need to separately kmalloc small ones
 };
 
-/* Filldir cache creation/deletion. */
+/* Cache creation/deletion routines. */
 void destroy_filldir_cache(void);
 int init_filldir_cache(void);
+int init_inode_cache(void);
+void destroy_inode_cache(void);
+int init_dentry_cache(void);
+void destroy_dentry_cache(void);
 
 /* Initialize and free readdir-specific  state. */
 int init_rdstate(struct file *file);
@@ -228,16 +238,16 @@ struct filldir_node *find_filldir_node(struct unionfs_dir_state *rdstate, const
 
 /* Directory hash table. */
 struct unionfs_dir_state {
-    unsigned int uds_cookie; /* The cookie, which is based off of uii_rdversion */
+	unsigned int uds_cookie; /* The cookie, which is based off of uii_rdversion */
 	unsigned int uds_offset; /* The entry we have returned. */
-    int uds_bindex;
+	int uds_bindex;
 	loff_t uds_dirpos; /* The offset within the lower level directory. */
-    int uds_size; /* How big is the hash table? */
-    int uds_hashentries; /* How many entries have been inserted? */
-    unsigned long uds_access;
-    /* This cache list is used when the inode keeps us around. */
-    struct list_head uds_cache;
-    struct list_head uds_list[0];
+	int uds_size; /* How big is the hash table? */
+	int uds_hashentries; /* How many entries have been inserted? */
+	unsigned long uds_access;
+	/* This cache list is used when the inode keeps us around. */
+	struct list_head uds_cache;
+	struct list_head uds_list[0];
 };
 /* We can only use 32-bits of offset for rdstate --- blech! */
 #define DIREOF (0xfffff)
@@ -259,11 +269,11 @@ static inline void verify_rdstate_offset(struct unionfs_dir_state *rdstate) {
 
 /* file private data. */
 struct unionfs_file_info {
-    struct file ** ufi_file;
-    int b_start;
-    int b_end;
-    atomic_t ufi_generation;
-    struct unionfs_dir_state *rdstate;
+	struct file ** ufi_file;
+	int b_start;
+	int b_end;
+	atomic_t ufi_generation;
+	struct unionfs_dir_state *rdstate;
 };
 
 /*
@@ -289,20 +299,27 @@ struct unionfs_file_info {
         fbend(file) = -1;\
      }
 /* Initialize array pointers for hidden files */
-#define init_ftohf_ptr(file, num) memset(ftohf_ptr(file), 0, (sizeof(file_t *) * num))
+#define init_ftohf_ptr(file, num) memset(ftohf_ptr(file), 0, (sizeof(struct file *) * num))
 
 // Inode TO Private Data
+#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,0)
 #define itopd(ino) ((struct unionfs_inode_info *)(ino)->u.generic_ip)
 #define itopd_lhs(ino) ((ino)->u.generic_ip)
+#else
+static inline struct unionfs_inode_info *itopd(const struct inode *inode) {
+	return &(container_of(inode, struct unionfs_inode_container, vfs_inode)->info);
+}
+#endif
+
 // Inode TO Hidden Inode
 // #define itohi(ino) itopd(ino)->uii_inode[itopd(ino)->b_start]
 #define itohi(ino) __itohi(ino, __FILE__, __FUNCTION__, __LINE__)
-static inline struct inode *__itohi(struct inode *ino, const char *file, const char *function, int line) {
-    struct inode *hidden_inode;
-    ASSERT2(itopd(ino)->b_start >= 0);
-    hidden_inode = itopd(ino)->uii_inode[itopd(ino)->b_start];
-    ASSERT2((atomic_read(&hidden_inode->i_count)) > 0);
-    return hidden_inode;
+static inline struct inode *__itohi(const struct inode *ino, const char *file, const char *function, int line) {
+	struct inode *hidden_inode;
+	ASSERT2(itopd(ino)->b_start >= 0);
+	hidden_inode = itopd(ino)->uii_inode[itopd(ino)->b_start];
+	ASSERT2((atomic_read(&hidden_inode->i_count)) > 0);
+	return hidden_inode;
 }
 
 // Inode To Hidden Inode based on index
@@ -315,7 +332,7 @@ static inline struct inode *__itohi(struct inode *ino, const char *file, const c
 #define ibend(ino) (itopd(ino)->b_end)
 // Initialize private data for inode
 #define init_itopd(ino)  \
-    {\
+    do {\
         PASSERT(ino->i_sb);\
         memset(itopd(ino), 0, sizeof(struct unionfs_inode_info));\
         itopd(ino)->b_start = -1; \
@@ -325,9 +342,9 @@ static inline struct inode *__itohi(struct inode *ino, const char *file, const c
 	itopd(ino)->uii_rdcount = 1;\
 	itopd(ino)->uii_hashsize = -1;\
 	INIT_LIST_HEAD(&itopd(ino)->uii_readdircache);\
-    }
+    } while(0)
 // Initialize array pointers for inode
-#define init_itohi_ptr(inode, num) memset(itohi_ptr(inode), 0, (sizeof(inode_t *) * num))
+#define init_itohi_ptr(inode, num) memset(itohi_ptr(inode), 0, (sizeof(struct inode *) * num))
 
 // Superblock TO Private Data
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
@@ -355,14 +372,15 @@ static inline struct inode *__itohi(struct inode *ino, const char *file, const c
 // Initialize private data for super block
 
 #define init_stopd(sb) \
-    {\
+    do {\
         memset(stopd(sb), 0, sizeof(struct unionfs_sb_info)); \
 	stopd(sb)->usi_thissb = sb; \
         stopd(sb)->b_start = -1;\
         stopd(sb)->b_end = -1;\
+        stopd(sb)->usi_diropaque = 1;\
 	atomic_set(&stopd(sb)->usi_generation, 1);\
 	init_alt_super(sb); \
-    }
+    } while (0)
 
 #ifdef SPLIT_VIEW_CACHES
 #define init_alt_super(sb) \
@@ -375,8 +393,8 @@ static inline struct inode *__itohi(struct inode *ino, const char *file, const c
 #endif
 
 // Initialize array pointers for hidden superblock
-#define init_stohs_ptr(sb, num) memset(stohs_ptr(sb), 0, (sizeof(super_block_t *) * num))
-#define init_stohiddenmnt_ptr(sb, num) memset(stohiddenmnt_ptr(sb), 0, (sizeof(vfs_mount_t *) * num))
+#define init_stohs_ptr(sb, num) memset(stohs_ptr(sb), 0, (sizeof(struct super_block *) * num))
+#define init_stohiddenmnt_ptr(sb, num) memset(stohiddenmnt_ptr(sb), 0, (sizeof(struct vfsmount *) * num))
 
 /* Get and put branches on the superblock. */
 #define branchget(sb, index) atomic_inc(&stopd(sb)->usi_sbcount[index]);
@@ -389,7 +407,7 @@ static inline struct inode *__itohi(struct inode *ino, const char *file, const c
 #define unlock_dpd2(dentry) __multilock_unlock(&dtopd(dentry)->udi_lock, file, function, line)
 
 // Dentry to Hidden Dentry Pointer
-#define dtohd_ptr(dentry) (dtopd(dentry)->udi_dentry)
+#define dtohd_ptr(dentry) (dtopd_nocheck(dentry)->udi_dentry)
 // Dentry TO Hidden Dentry (and private data --- for the future)
 #define dtopd(dent) __dtopd(dent, 1, __FILE__, __FUNCTION__, __LINE__)
 static inline struct unionfs_dentry_info *__dtopd(const struct dentry *dent, int check, const char *file, const char *function, int line) {
@@ -402,7 +420,7 @@ static inline struct unionfs_dentry_info *__dtopd(const struct dentry *dent, int
 		PASSERT2(ret);
 		if (check) {
 			if ((ret->udi_bend > ret->udi_bcount) || (ret->udi_bend > sbmax(dent->d_sb))) {
-				printk("udi_bend = %d, udi_count = %d, sbmax = %d\n", ret->udi_bend, ret->udi_bcount, sbmax(dent->d_sb));
+				printk(KERN_EMERG "udi_bend = %d, udi_count = %d, sbmax = %d\n", ret->udi_bend, ret->udi_bcount, sbmax(dent->d_sb));
 			}
 			ASSERT2(ret->udi_bend <= ret->udi_bcount);
 			ASSERT2(ret->udi_bend <= sbmax(dent->d_sb));
@@ -418,190 +436,197 @@ static inline struct unionfs_dentry_info *__dtopd(const struct dentry *dent, int
 // Dentry to Hidden dentry
 // #define dtohd(dent) dtopd(dent)->udi_dentry[dtopd(dent)->b_start]
 #define dtohd(dent) __dtohd(dent, __FILE__, __FUNCTION__, __LINE__)
-static inline struct dentry *__dtohd(struct dentry *dent, const char *file, const char *function, int line) {
-    struct dentry *d;
+static inline struct dentry *__dtohd(const struct dentry *dent, const char *file, const char *function, int line) {
+	struct dentry *d;
 
-    lock_dpd2(dent);
+	lock_dpd2(dent);
 
-    PASSERT2(dent);
-    PASSERT2(dent->d_op);
-    ASSERT2(dent->d_op == &unionfs_dops);
-    ASSERT2(dent->d_sb->s_op == &unionfs_sops);
-    if (dent->d_inode) {
-	PASSERT2(dent->d_inode);
-	ASSERT(dent->d_inode->i_op == &unionfs_main_iops ||
-		dent->d_inode->i_op == &unionfs_dir_iops ||
-		dent->d_inode->i_op == &unionfs_symlink_iops);
-    }
+	PASSERT2(dent);
+	PASSERT2(dent->d_op);
+	ASSERT2(dent->d_op == &unionfs_dops);
+	ASSERT2(dent->d_sb->s_op == &unionfs_sops);
+	if (dent->d_inode) {
+		PASSERT2(dent->d_inode);
+		ASSERT(dent->d_inode->i_op == &unionfs_main_iops ||
+				dent->d_inode->i_op == &unionfs_dir_iops ||
+				dent->d_inode->i_op == &unionfs_symlink_iops);
+	}
 
-    ASSERT2(dtopd(dent)->udi_bstart >= 0);
-    d = dtopd(dent)->udi_dentry[dtopd(dent)->udi_bstart];
-    if (d) {
-	PASSERT2(d);
-    	ASSERT2((atomic_read(&d->d_count)) > 0);
-    }
+	ASSERT2(dtopd(dent)->udi_bstart >= 0);
+	d = dtopd(dent)->udi_dentry[dtopd(dent)->udi_bstart];
+	if (d) {
+		PASSERT2(d);
+		ASSERT2((atomic_read(&d->d_count)) > 0);
+	}
 
-    unlock_dpd2(dent);
+	unlock_dpd2(dent);
 
-    return d;
+	return d;
 }
 
 // Dentry to Hidden Dentry based on index
-#define set_dtohd_index(dent, index, val) __set_dtohd_index(dent, index, val, __FILE__, __FUNCTION__, __LINE__)
-static inline struct dentry *__set_dtohd_index(struct dentry *dent, int index, struct dentry *val, const char *file, const char *function, int line) {
-    PASSERT2(dent);
+#define set_dtohd_index(dent, index, val) __set_dtohd_index(dent, index, val, __FILE__, __FUNCTION__, __LINE__, 1)
+#define set_dtohd_index_nocheck(dent, index, val) __set_dtohd_index(dent, index, val, __FILE__, __FUNCTION__, __LINE__, 0)
+static inline struct dentry *__set_dtohd_index(struct dentry *dent, int index, struct dentry *val, const char *file, const char *function, int line, int docheck) {
+	PASSERT2(dent);
 
-    lock_dpd2(dent);
+	lock_dpd2(dent);
 
-    PASSERT2(dent->d_sb);
-    PASSERT2(dtopd(dent));
-    PASSERT2(dtopd(dent)->udi_dentry);
-    ASSERT2(index >= 0);
-    if (index > sbend(dent->d_sb)) {
-	printk("Dentry index out of super bounds: index=%d, sbend=%d\n", index, sbend(dent->d_sb));
-        ASSERT2(index <= sbend(dent->d_sb));
-    }
-    if (index > dtopd(dent)->udi_bcount) {
-	printk("Dentry index out of array bounds: index=%d, count=%d\n", index, dtopd(dent)->udi_bcount);
-	printk("Generation of dentry: %d\n", atomic_read(&dtopd(dent)->udi_generation));
-	printk("Generation of sb: %d\n", atomic_read(&stopd(dent->d_sb)->usi_generation));
-        ASSERT2(index <= dtopd(dent)->udi_bcount);
-    }
-    if (val) {
-	PASSERT2(val);
-    }
-    dtopd(dent)->udi_dentry[index] = val;
+	PASSERT2(dent->d_sb);
+	PASSERT2(dtopd(dent));
+	PASSERT2(dtopd(dent)->udi_dentry);
+	ASSERT2(index >= 0);
+	if(docheck)
+	{
+		if (index > sbend(dent->d_sb)) {
+			printk("Dentry index out of super bounds: index=%d, sbend=%d\n", index, sbend(dent->d_sb));
+			ASSERT2(index <= sbend(dent->d_sb));
+		}
+		if (index > dtopd(dent)->udi_bcount) {
+			printk("Dentry index out of array bounds: index=%d, count=%d\n", index, dtopd(dent)->udi_bcount);
+			printk("Generation of dentry: %d\n", atomic_read(&dtopd(dent)->udi_generation));
+			printk("Generation of sb: %d\n", atomic_read(&stopd(dent->d_sb)->usi_generation));
+			ASSERT2(index <= dtopd(dent)->udi_bcount);
+		}
+	}
+	if (val) {
+		PASSERT2(val);
+	}
+	dtopd(dent)->udi_dentry[index] = val;
 
-    unlock_dpd2(dent);
-    return val;
+	unlock_dpd2(dent);
+	return val;
 }
 
-#define dtohd_index(dent, index) __dtohd_index(dent, index, __FILE__, __FUNCTION__, __LINE__)
-static inline struct dentry *__dtohd_index(const struct dentry *dent, int index, const char *file, const char *function, int line) {
-    struct dentry *d;
+#define dtohd_index(dent, index) __dtohd_index(dent, index, __FILE__, __FUNCTION__, __LINE__,1)
+#define dtohd_index_nocheck(dent, index) __dtohd_index(dent, index, __FILE__, __FUNCTION__, __LINE__,0)
+static inline struct dentry *__dtohd_index(const struct dentry *dent, int index, const char *file, const char *function, int line,int docheck) {
+	struct dentry *d;
 
-    PASSERT2(dent);
-    lock_dpd2(dent);
-    PASSERT2(dent->d_sb);
-    PASSERT2(dtopd(dent));
-    PASSERT2(dtopd(dent)->udi_dentry);
-    ASSERT2(index >= 0);
-    if (index > sbend(dent->d_sb)) {
-	printk("Dentry index out of super bounds: index=%d, sbend=%d\n", index, sbend(dent->d_sb));
-        ASSERT2(index <= sbend(dent->d_sb));
-    }
-    if (index > dtopd(dent)->udi_bcount) {
-	printk("Dentry index out of array bounds: index=%d, count=%d\n", index, dtopd(dent)->udi_bcount);
-	printk("Generation of dentry: %d\n", atomic_read(&dtopd(dent)->udi_generation));
-	printk("Generation of sb: %d\n", atomic_read(&stopd(dent->d_sb)->usi_generation));
-        ASSERT2(index <= dtopd(dent)->udi_bcount);
-    }
-    d = dtopd(dent)->udi_dentry[index];
-    if (d) {
-	PASSERT2(d);
-    	ASSERT2((atomic_read(&d->d_count)) > 0);
-    }
-    unlock_dpd2(dent);
-    return d;
+	PASSERT2(dent);
+	lock_dpd2(dent);
+	PASSERT2(dent->d_sb);
+	PASSERT2(dtopd(dent));
+	PASSERT2(dtopd(dent)->udi_dentry);
+	ASSERT2(index >= 0);
+	if(docheck)
+	{
+		if (index > sbend(dent->d_sb)) {
+			printk("Dentry index out of super bounds: index=%d, sbend=%d\n", index, sbend(dent->d_sb));
+			ASSERT2(index <= sbend(dent->d_sb));
+		}
+		if (index > dtopd(dent)->udi_bcount) {
+			printk("Dentry index out of array bounds: index=%d, count=%d\n", index, dtopd(dent)->udi_bcount);
+			printk("Generation of dentry: %d\n", atomic_read(&dtopd(dent)->udi_generation));
+			printk("Generation of sb: %d\n", atomic_read(&stopd(dent->d_sb)->usi_generation));
+			ASSERT2(index <= dtopd(dent)->udi_bcount);
+		}
+	}
+	d = dtopd(dent)->udi_dentry[index];
+	if (d) {
+		PASSERT2(d);
+		ASSERT2((atomic_read(&d->d_count)) > 0);
+	}
+	unlock_dpd2(dent);
+	return d;
 }
 // Macros to refer to branch indices
 
 #define dbend(dentry) __dbend(dentry, __FILE__, __FUNCTION__, __LINE__)
 static inline int __dbend(const struct dentry *dentry, const char *file, const char *function, int line) {
-    return dtopd(dentry)->udi_bend;
+	return dtopd(dentry)->udi_bend;
 }
 #define set_dbend(dentry, val) __set_dbend(dentry, val, __FILE__, __FUNCTION__, __LINE__)
 static inline int __set_dbend(const struct dentry *dentry, int val, const char *file, const char *function, int line) {
-    if (val < 0) {
-	ASSERT2(val == -1);
-    }
-    ASSERT2(val <= __dtopd(dentry, 0, file, function, line)->udi_bcount);
-    ASSERT2(val <= sbmax(dentry->d_sb));
-    dtopd(dentry)->udi_bend = val;
-    return __dtopd(dentry, 1, file, function, line)->udi_bend;
+	if (val < 0) {
+		ASSERT2(val == -1);
+	}
+	ASSERT2(val <= __dtopd(dentry, 0, file, function, line)->udi_bcount);
+	ASSERT2(val <= sbmax(dentry->d_sb));
+	dtopd(dentry)->udi_bend = val;
+	return __dtopd(dentry, 1, file, function, line)->udi_bend;
 }
 
 
 #define dbstart(dentry) __dbstart(dentry, __FILE__, __FUNCTION__, __LINE__)
 static inline int __dbstart(const struct dentry *dentry, const char *file, const char *function, int line) {
-    return dtopd(dentry)->udi_bstart;
+	return dtopd(dentry)->udi_bstart;
 }
 #define set_dbstart(dentry, val) __set_dbstart(dentry, val, __FILE__, __FUNCTION__, __LINE__)
 static inline int __set_dbstart(const struct dentry *dentry, int val, const char *file, const char *function, int line) {
-    if (val < 0) {
-	ASSERT2(val == -1);
-    }
-    ASSERT2(val <= __dtopd(dentry, 0, file, function, line)->udi_bcount);
-    ASSERT2(val <= sbmax(dentry->d_sb));
-    __dtopd(dentry, 0, file, function, line)->udi_bstart = val;
-    return __dtopd(dentry, 1, file, function, line)->udi_bstart;
+	if (val < 0) {
+		ASSERT2(val == -1);
+	}
+	ASSERT2(val <= __dtopd(dentry, 0, file, function, line)->udi_bcount);
+	ASSERT2(val <= sbmax(dentry->d_sb));
+	__dtopd(dentry, 0, file, function, line)->udi_bstart = val;
+	return __dtopd(dentry, 1, file, function, line)->udi_bstart;
 }
 
 
-#define init_dtopd(dentry, sb) \
-    {\
-	multilock_init(&dtopd_nocheck(dentry)->udi_lock); \
-	reinit_dtopd(dentry, sb); \
-    }
-
-#define reinit_dtopd(dentry, sb) {\
-	dtopd_nocheck(dentry)->udi_dentry = NULL;\
-        dtopd_nocheck(dentry)->udi_bstart = -1;\
-        dtopd_nocheck(dentry)->udi_bend = -1;\
-        dtopd_nocheck(dentry)->udi_bcount = 0;\
-        atomic_set(&dtopd_nocheck(dentry)->udi_generation, atomic_read(&stopd(sb)->usi_generation));\
-    }
-#define init_dentry_info(ptr) { \
-        memset(ptr, 0, sizeof(struct unionfs_dentry_info));\
-        ptr->udi_bstart = -1;\
-        ptr->udi_bend = -1;\
+#define dbopaque(dentry) __dbopaque(dentry, __FILE__, __FUNCTION__, __LINE__)
+static inline int __dbopaque(const struct dentry *dentry, const char *file, const char *function, int line) {
+	return dtopd(dentry)->udi_bopaque;
+}
+#define set_dbopaque(dentry, val) __set_dbopaque(dentry, val, __FILE__, __FUNCTION__, __LINE__)
+static inline int __set_dbopaque(const struct dentry *dentry, int val, const char *file, const char *function, int line) {
+	if (val < 0) {
+		ASSERT2(val == -1);
+	}
+	ASSERT2(val <= __dtopd(dentry, 0, file, function, line)->udi_bcount);
+	ASSERT2(val <= sbmax(dentry->d_sb));
+	ASSERT2(val <= dbend(dentry));
+	__dtopd(dentry, 0, file, function, line)->udi_bopaque = val;
+	return __dtopd(dentry, 1, file, function, line)->udi_bopaque;
 }
-// Initialize array pointers for sb
-#define init_dtohd_ptr(dent, num) do { \
-	struct dentry *d = dent; \
-	dtopd(d)->udi_bcount = num; \
-	memset(dtohd_ptr(d), 0, (sizeof(dentry_t *) * num)); \
-} while(0)
 
+extern int new_dentry_private_data(struct dentry *dentry);
+void free_dentry_private_data(struct unionfs_dentry_info *udi);
 #define sbt(sb) ((sb)->s_type->name)
 
 /*
  * EXTERNALS:
  */
 /* replicates the directory structure upto given dentry in given branch */
-extern dentry_t *unionfs_create_dirs(inode_t *dir, dentry_t *dentry, int bindex);
+extern struct dentry *unionfs_create_dirs(struct inode *dir, struct dentry *dentry, int bindex);
 
 /* partial lookup */
-extern int unionfs_partial_lookup(dentry_t *dentry);
+extern int unionfs_partial_lookup(struct dentry *dentry);
 
 /* Pass an unionfs dentry and an index and it will try to create a whiteout in branch 'index'.
    On error, it will proceed to a branch to the left */
-extern int create_whiteout(dentry_t *dentry, int start);
-extern int create_whiteout_parent(dentry_t *parent_dentry, const char *filename, int start);
+extern int create_whiteout(struct dentry *dentry, int start);
+extern int create_whiteout_parent(struct dentry *parent_dentry, const char *filename, int start);
 /* copies a file from dbstart to newbindex branch */
-extern int unionfs_copyup_file(inode_t *dir, file_t *file, int bstart, int newbindex, int size);
+extern int unionfs_copyup_file(struct inode *dir, struct file *file, int bstart, int newbindex, int size);
 
 /* copies a dentry from dbstart to newbindex branch */
-extern int unionfs_copyup_dentry_len(inode_t *dir, dentry_t *dentry, int bstart, int new_bindex, file_t **copyup_file, int len);
+extern int unionfs_copyup_dentry_len(struct inode *dir, struct dentry *dentry, int bstart, int new_bindex, struct file **copyup_file, int len);
 
-extern int create_dir_whs(dentry_t *dentry, int cur_index) ;
+extern int create_dir_whs(struct dentry *dentry, int cur_index) ;
 
-extern int remove_whiteouts(dentry_t *dentry, dentry_t *hidden_dentry, int bindex);
+extern int remove_whiteouts(struct dentry *dentry, struct dentry *hidden_dentry, int bindex);
 
 /* Is this directory empty: 0 if it is empty, -ENOTEMPTY if not. */
-extern int check_empty(dentry_t *dentry, struct unionfs_dir_state **namelist);
+extern int check_empty(struct dentry *dentry, struct unionfs_dir_state **namelist);
 /* Delete whiteouts from this directory in branch bindex. */
 extern int delete_whiteouts(struct dentry *dentry, int bindex, struct unionfs_dir_state *namelist);
 
 /* Re-lookup a hidden dentry. */
 extern int unionfs_refresh_hidden_dentry(struct dentry *dentry, int bindex);
 
-extern void unionfs_reinterpose(dentry_t *this_dentry);
-extern struct super_block *unionfs_duplicate_super(super_block_t *sb);
+extern void unionfs_reinterpose(struct dentry *this_dentry);
+extern struct super_block *unionfs_duplicate_super(struct super_block *sb);
 
 /* Locking functions. */
-extern int unionfs_setlk(file_t *file, int cmd, struct file_lock *fl);
-extern int unionfs_getlk(file_t *file, struct file_lock *fl);
+extern int unionfs_setlk(struct file *file, int cmd, struct file_lock *fl);
+extern int unionfs_getlk(struct file *file, struct file_lock *fl);
+
+/* Inode operations */
+extern int unionfs_rename(struct inode *old_dir, struct dentry *old_dentry,
+	       struct inode *new_dir, struct dentry *new_dentry);
+int unionfs_unlink(struct inode *dir, struct dentry *dentry);
+int unionfs_rmdir(struct inode *dir, struct dentry *dentry);
 
 /* The values for unionfs_interpose's flag. */
 #define INTERPOSE_DEFAULT	0
@@ -609,16 +634,18 @@ extern int unionfs_getlk(file_t *file, struct file_lock *fl);
 #define INTERPOSE_REVAL		2
 #define INTERPOSE_REVAL_NEG	3
 #define INTERPOSE_LINK		4
+#define INTERPOSE_PARTIAL	5
+#define INTERPOSE_PARTIAL_NEG	6
 
-extern int unionfs_interpose(dentry_t *this_dentry, super_block_t *sb, int flag);
+extern int unionfs_interpose(struct dentry *this_dentry, struct super_block *sb, int flag);
 
 /* Branch management ioctls. */
-int unionfs_ioctl_branchcount(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg);
-int unionfs_ioctl_incgen(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg);
-int unionfs_ioctl_addbranch(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg);
-int unionfs_ioctl_delbranch(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg);
-int unionfs_ioctl_rdwrbranch(inode_t *inode, file_t *unused_file, unsigned int cmd, unsigned long arg);
-int unionfs_ioctl_superduper(inode_t *inode, file_t *file, unsigned int cmd, unsigned long arg);
+int unionfs_ioctl_branchcount(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
+int unionfs_ioctl_incgen(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
+int unionfs_ioctl_addbranch(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
+int unionfs_ioctl_delbranch(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
+int unionfs_ioctl_rdwrbranch(struct inode *inode, struct file *unused_file, unsigned int cmd, unsigned long arg);
+int unionfs_ioctl_superduper(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
 
 
 #if defined(UNIONFS_XATTR) && LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,20)
@@ -630,7 +657,7 @@ extern int unionfs_getxattr(struct dentry *dentry, const char *name, void *value
 extern int unionfs_removexattr(struct dentry *dentry, const char *name);
 extern int unionfs_listxattr(struct dentry *dentry, char *list, size_t size);
 
-#if defined(FIST_SETXATTR_CONSTVOID) || (LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0))
+#if defined(FIST_SETXATTR_CONSTVOID) || (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
 int unionfs_setxattr(struct dentry *dentry, const char *name, const void *value, size_t size, int flags);
 #else
 int unionfs_setxattr(struct dentry *dentry, const char *name, void *value, size_t size, int flags);
@@ -643,85 +670,80 @@ int unionfs_setxattr(struct dentry *dentry, const char *name, void *value, size_
     dst->i_blocks = src->i_blocks;
 
 /* returns the sum of the n_link values of all the underlying inodes of the passed inode */
-static inline int
-get_nlinks(inode_t *inode)
+static inline int get_nlinks(struct inode *inode)
 {
-    int sum_nlinks = 2, bindex;
-    inode_t *hidden_inode;
+	int sum_nlinks = 2, bindex;
+	struct inode *hidden_inode;
 
-    PASSERT(inode);
-    if (!S_ISDIR(inode->i_mode)) {
-	return itohi(inode)->i_nlink;
-    }
-
-    for (bindex = ibstart(inode); bindex <= ibend(inode); bindex++) {
-        hidden_inode = itohi_index(inode, bindex);
-	if (!hidden_inode || !S_ISDIR(hidden_inode->i_mode)) {
-	    continue;
+	PASSERT(inode);
+	if (!S_ISDIR(inode->i_mode)) {
+		return itohi(inode)->i_nlink;
 	}
-	sum_nlinks += (hidden_inode->i_nlink - 2);
-    }
-    return sum_nlinks;
+
+	for (bindex = ibstart(inode); bindex <= ibend(inode); bindex++) {
+		hidden_inode = itohi_index(inode, bindex);
+		if (!hidden_inode || !S_ISDIR(hidden_inode->i_mode)) {
+			continue;
+		}
+		sum_nlinks += (hidden_inode->i_nlink - 2);
+	}
+	return sum_nlinks;
 }
 
-static inline void
-fist_copy_attr_atime(inode_t *dest, const inode_t *src)
+static inline void fist_copy_attr_atime(struct inode *dest, const struct inode *src)
 {
-    PASSERT(dest);
-    PASSERT(src);
-    dest->i_atime = src->i_atime;
+	PASSERT(dest);
+	PASSERT(src);
+	dest->i_atime = src->i_atime;
 }
-static inline void
-fist_copy_attr_times(inode_t *dest, const inode_t *src)
+static inline void fist_copy_attr_times(struct inode *dest, const struct inode *src)
 {
-    PASSERT(dest);
-    PASSERT(src);
-    dest->i_atime = src->i_atime;
-    dest->i_mtime = src->i_mtime;
-    dest->i_ctime = src->i_ctime;
+	PASSERT(dest);
+	PASSERT(src);
+	dest->i_atime = src->i_atime;
+	dest->i_mtime = src->i_mtime;
+	dest->i_ctime = src->i_ctime;
 }
-static inline void
-fist_copy_attr_timesizes(inode_t *dest, const inode_t *src)
+static inline void fist_copy_attr_timesizes(struct inode *dest, const struct inode *src)
 {
-    PASSERT(dest);
-    PASSERT(src);
-    dest->i_atime = src->i_atime;
-    dest->i_mtime = src->i_mtime;
-    dest->i_ctime = src->i_ctime;
-    copy_inode_size(dest, src);
+	PASSERT(dest);
+	PASSERT(src);
+	dest->i_atime = src->i_atime;
+	dest->i_mtime = src->i_mtime;
+	dest->i_ctime = src->i_ctime;
+	copy_inode_size(dest, src);
 }
-static inline void
-fist_copy_attr_all(inode_t *dest, const inode_t *src)
+static inline void fist_copy_attr_all(struct inode *dest, const struct inode *src)
 {
 
-    print_entry_location();
-    PASSERT(dest);
-    PASSERT(src);
-    dest->i_mode = src->i_mode;
-    dest->i_nlink = get_nlinks(dest);
-    dest->i_uid = src->i_uid;
-    dest->i_gid = src->i_gid;
-    dest->i_rdev = src->i_rdev;
-    dest->i_atime = src->i_atime;
-    dest->i_mtime = src->i_mtime;
-    dest->i_ctime = src->i_ctime;
-    dest->i_blksize = src->i_blksize;
-    dest->i_blkbits = src->i_blkbits;
-    copy_inode_size(dest, src);
+	print_entry_location();
+	PASSERT(dest);
+	PASSERT(src);
+	dest->i_mode = src->i_mode;
+	dest->i_nlink = get_nlinks(dest);
+	dest->i_uid = src->i_uid;
+	dest->i_gid = src->i_gid;
+	dest->i_rdev = src->i_rdev;
+	dest->i_atime = src->i_atime;
+	dest->i_mtime = src->i_mtime;
+	dest->i_ctime = src->i_ctime;
+	dest->i_blksize = src->i_blksize;
+	dest->i_blkbits = src->i_blkbits;
+	copy_inode_size(dest, src);
 
-//DQ: This was a change I noticed in the templates. In 2.6 they removedi_attr_flags.
-//Which makes me think they rolled it into flags.
+	//DQ: This was a change I noticed in the templates. In 2.6 they removedi_attr_flags.
+	//Which makes me think they rolled it into flags.
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
-    dest->i_attr_flags = src->i_attr_flags;
+	dest->i_attr_flags = src->i_attr_flags;
 #else
-    dest->i_flags = src->i_flags;
+	dest->i_flags = src->i_flags;
 #endif
 
-    print_exit_location();
+	print_exit_location();
 }
 
 
-dentry_t *unionfs_lookup_backend(inode_t *dir, dentry_t *dentry, int interpose_flag);
+struct dentry *unionfs_lookup_backend(struct dentry *dentry, int lookupmode);
 int is_stale_inode(struct inode * inode);
 void make_stale_inode(struct inode * inode);
 
@@ -731,69 +753,74 @@ void make_stale_inode(struct inode * inode);
 #define IS_COPYUP_ERR(err) (err == -EROFS)
 
 /* unionfs_open, check if we need to copyup the file */
-#define IS_WRITE_FLAG(flag) (flag & (O_WRONLY | O_RDWR | O_APPEND))
+#define OPEN_WRITE_FLAGS (O_WRONLY | O_RDWR | O_APPEND)
+#define IS_WRITE_FLAG(flag) (flag & (OPEN_WRITE_FLAGS))
 
 /* Is this file on a read-only branch? */
 static inline int __is_robranch_super(struct super_block *sb, int index, char *file, const char *function, int line) {
-    int err = 0;
+	int err = 0;
 
-    print_util_entry_location();
+	print_util_entry_location();
 
-    ASSERT(index >= 0);
-    PASSERT(sb);
-    PASSERT(stopd(sb));
-    PASSERT(stopd(sb)->usi_branchperms);
+	ASSERT(index >= 0);
+	PASSERT(sb);
+	PASSERT(stopd(sb));
+	PASSERT(stopd(sb)->usi_branchperms);
 
-    if (!(stopd(sb)->usi_branchperms[index] & MAY_WRITE)) {
-	err = -EROFS;
-    }
-    print_util_exit_status(err);
-    return err;
+	if (!(stopd(sb)->usi_branchperms[index] & MAY_WRITE)) {
+		err = -EROFS;
+	}
+	print_util_exit_status(err);
+	return err;
 }
 
 /* Is this file on a read-only branch? */
 static inline int __is_robranch_index(struct dentry *dentry, int index, char *file, const char *function, int line) {
-    int err = 0;
+	int err = 0;
 
-    print_util_entry_location();
+	print_util_entry_location();
 
-    PASSERT2(dentry);
-    ASSERT2(index >= 0);
-    PASSERT2(dtohd_index(dentry, index));
-    PASSERT2(dtohd_index(dentry, index)->d_inode);
+	PASSERT2(dentry);
+	ASSERT2(index >= 0);
+	PASSERT2(dtohd_index(dentry, index));
+	PASSERT2(dtohd_index(dentry, index)->d_inode);
 
-    if (!(stopd(dentry->d_sb)->usi_branchperms[index] & MAY_WRITE)) {
-	err = -EROFS;
-    } else if (IS_RDONLY(dtohd_index(dentry, index)->d_inode)) {
-	err = -EROFS;
-    }
+	if (!(stopd(dentry->d_sb)->usi_branchperms[index] & MAY_WRITE)) {
+		err = -EROFS;
+	} else if (IS_RDONLY(dtohd_index(dentry, index)->d_inode)) {
+		err = -EROFS;
+	}
 
-    print_util_exit_status(err);
+	print_util_exit_status(err);
 
-    return err;
+	return err;
 }
 static inline int __is_robranch(struct dentry *dentry, char *file, const char *function, int line) {
-    int index;
-    int err;
+	int index;
+	int err;
 
-    print_util_entry_location();
+	print_util_entry_location();
 
-    PASSERT2(dentry);
-    PASSERT2(dtopd(dentry));
-    index = dtopd(dentry)->udi_bstart;
-    ASSERT2(index >= 0);
+	PASSERT2(dentry);
+	PASSERT2(dtopd(dentry));
+	index = dtopd(dentry)->udi_bstart;
+	ASSERT2(index >= 0);
 
-    err =  __is_robranch_index(dentry, index, file, function, line);
+	err =  __is_robranch_index(dentry, index, file, function, line);
 
-    print_util_exit_status(err);
+	print_util_exit_status(err);
 
-    return err;
+	return err;
 }
 
 #define is_robranch(d) __is_robranch(d, __FILE__, __FUNCTION__, __LINE__)
 #define is_robranch_super(s, n) __is_robranch_super(s, n, __FILE__, __FUNCTION__, __LINE__)
 
-/* auto-generated extern definitions */
+/* If a directory contains this file, then it is opaque.  We start with the
+ * .wh. flag so that it is blocked by loomkup.
+ */
+#define UNIONFS_DIR_OPAQUE_NAME "__dir_opaque"
+#define UNIONFS_DIR_OPAQUE ".wh." UNIONFS_DIR_OPAQUE_NAME
 
 #endif /* __KERNEL__ */
 
@@ -824,13 +851,23 @@ static inline int __is_robranch(struct dentry *dentry, char *file, const char *f
 
 #define VALID_MOUNT_FLAGS (GLOBAL_ERR_PASSUP | DELETE_FIRST | DELETE_WHITEOUT | COPYUP_OWNER | COPYUP_FS_MOUNTER | SETATTR_ALL)
 
+/*
+ * Defines and structs for persistant inodes
+ */
+
+#define P_CREAT 0x01
+
+#define UNIONFS_MAP_FILE_NAME "__unionfs_inode_map"
+#define UNIOFNS_MAP_FILE ".wh." UNIONFS_MAP_FILE_NAME
+
 /* ioctls */
 
 #define EUNIONFS_NO_WHITEOUT   1025
 #endif	/* not __UNIONFS_H_ */
 
 /*
+ * vim:shiftwidth=8
  * Local variables:
- * c-basic-offset: 4
+ * c-basic-offset: 8
  * End:
  */
